@@ -3,10 +3,10 @@ from typing import Iterator, List, Union
 
 from Bio.SeqFeature import FeatureLocation, CompoundLocation
 
-from inscripta.biocantor.exc import NoSuchAncestorException
+from inscripta.biocantor.exc import NoSuchAncestorException, NullParentException
 from inscripta.biocantor.location.distance import DistanceType
 from inscripta.biocantor.location.strand import Strand
-from inscripta.biocantor.parent import Parent
+from inscripta.biocantor.parent import Parent, SequenceType
 from inscripta.biocantor.sequence import Sequence
 from inscripta.biocantor.util.object_validation import ObjectValidation
 
@@ -14,10 +14,10 @@ from inscripta.biocantor.util.object_validation import ObjectValidation
 class Location(ABC):
     """Abstract location with respect to a coordinate system"""
 
-    @abstractmethod
     def __len__(self):
         """Returns the length (number of positions) of this Location. For subclasses representing discontiguous
         locations, regions between blocks are not considered."""
+        return self.length
 
     @abstractmethod
     def __str__(self):
@@ -27,14 +27,29 @@ class Location(ABC):
     def __eq__(self, other):
         """Returns True iff this Location is equal to other object"""
 
+    @abstractmethod
+    def __hash__(self):
+        """Returns a hash code satisfying location1 == location2 => hash(location1) == hash(location2)"""
+
     @abstractmethod
     def __repr__(self):
         """Returns the 'official' string representation of this Location"""
 
-    @property
-    @abstractmethod
-    def parent(self) -> Parent:
-        """Returns the parent of this Location"""
+    # The 0-based start position of this Location on its parent
+    start: int
+
+    # The 0-based exclusive end position of this Location on its parent
+    end: int
+
+    # The strand of this Location with respect to its parent
+    strand: Strand
+
+    # The parent of this Location
+    parent: Parent
+
+    # The length (number of positions) of this Location. For subclasses representing discontiguous locations,
+    # regions between blocks are not considered
+    length: int
 
     @property
     def parent_id(self) -> str:
@@ -46,21 +61,6 @@ class Location(ABC):
         """Returns the sequence type of the parent"""
         return self.parent.sequence_type if self.parent else None
 
-    @property
-    @abstractmethod
-    def strand(self) -> Strand:
-        """Returns the strand of this Location with respect to its parent"""
-
-    @property
-    @abstractmethod
-    def start(self) -> int:
-        """Returns the 0-based start position of this Location on its parent"""
-
-    @property
-    @abstractmethod
-    def end(self) -> int:
-        """Returns the 0-based exclusive end position of this Location on its parent"""
-
     @property
     @abstractmethod
     def is_contiguous(self) -> bool:
@@ -81,6 +81,11 @@ class Location(ABC):
     def is_overlapping(self) -> bool:
         """Returns True if this interval contains overlaps; always False for SingleInterval"""
 
+    @property
+    @abstractmethod
+    def _full_span_interval(self) -> "Location":
+        """Returns the full span of this interval; is trivial for a SingleInterval and EmptyLocation"""
+
     @abstractmethod
     def scan_blocks(self) -> Iterator["Location"]:
         """Returns an iterator over blocks in order relative to strand of this Location"""
@@ -118,36 +123,45 @@ class Location(ABC):
     def relative_to_parent_pos(self, relative_pos: int) -> int:
         """Converts a position relative to this Location to a position on the parent"""
 
-    def parent_to_relative_location(self, parent_location: "Location") -> "Location":
+    def parent_to_relative_location(self, parent_location: "Location", optimize_blocks: bool = True) -> "Location":
         """Converts a Location on the parent to a Location relative to this Location.
 
         Parameters
         ----------
         parent_location
             Location with the same parent as this Location. Both parents can be None.
+        optimize_blocks
+            Run optimize_blocks on the resulting location?
 
         Returns
         -------
         New Location relative to this Location.
         """
-        return parent_location.location_relative_to(self)
+        return parent_location.location_relative_to(self, optimize_blocks=optimize_blocks)
 
-    def location_relative_to(self, other: "Location") -> "Location":
+    def location_relative_to(self, other: "Location", optimize_blocks: bool = True) -> "Location":
         """Converts this Location to a Location relative to another Location. The Locations must overlap.
-        The returned value represents the relative location of the overlap within the other Location."""
+        The returned value represents the relative location of the overlap within the other Location.
+
+        If ``optimize_blocks`` is ``True``, the resulting Location will not have any adjacent or overlapping
+        intervals. This is often desirable, because the output of this function can have weird coordinates
+        when the locations are overlapping or adjacent. However, there are some cases where it is desirable
+        to retain the original block structure. One such example are CDS where adjacent blocks or overlapping
+        blocks are used to model frameshifts or indels.
+        """
         if other.is_empty:
             return other
         if self.parent or other.parent:
             if not self.parent and other.parent:
-                raise ValueError(
+                raise NullParentException(
                     "Parents must be both null or both non-null:\n{}\n  !=\n{}".format(self.parent, other.parent)
                 )
             ObjectValidation.require_parents_equal_except_location(self.parent, other.parent)
         ObjectValidation.require_locations_overlap(self, other)
-        return self._location_relative_to(other)
+        return self._location_relative_to(other, optimize_blocks=optimize_blocks)
 
     @abstractmethod
-    def _location_relative_to(self, other: "Location") -> "Location":
+    def _location_relative_to(self, other: "Location", optimize_blocks: bool = True) -> "Location":
         raise NotImplementedError
 
     @abstractmethod
@@ -202,7 +216,13 @@ class Location(ABC):
             yield self.relative_interval_to_parent_location(curr_start, curr_start + window_size, Strand.PLUS)
 
     @abstractmethod
-    def has_overlap(self, other: "Location", match_strand: bool = False) -> bool:
+    def has_overlap(
+        self,
+        other: "Location",
+        match_strand: bool = False,
+        full_span: bool = False,
+        strict_parent_compare: bool = False,
+    ) -> bool:
         """Returns True iff this Location shares at least one position with the given Location.
         For subclasses representing discontiguous locations, regions between blocks are not considered.
 
@@ -212,6 +232,10 @@ class Location(ABC):
             Other Location
         match_strand
             If set to True, automatically return False if given interval Strand does not match this Location's Strand
+        full_span
+            If set to True, compare the full span of this Location to the full span of the other Location.
+        strict_parent_compare
+            Raise MismatchedParentException if parents do not match
 
         Returns
         -------
@@ -262,7 +286,7 @@ class Location(ABC):
         """Returns a BioPython interval type; since they do not have a shared base class, we need a union"""
         pass
 
-    def first_ancestor_of_type(self, sequence_type: str) -> Parent:
+    def first_ancestor_of_type(self, sequence_type: Union[str, SequenceType]) -> Parent:
         """Returns the Parent object representing the closest ancestor (parent, parent of parent, etc.)
         of this location which has the given sequence type. Raises NoSuchAncestorException if no ancestor with
         the given type exists."""
@@ -270,14 +294,14 @@ class Location(ABC):
             raise NoSuchAncestorException("Location has no parent")
         return self.parent.first_ancestor_of_type(sequence_type, include_self=True)
 
-    def has_ancestor_of_type(self, sequence_type: str) -> bool:
+    def has_ancestor_of_type(self, sequence_type: Union[str, SequenceType]) -> bool:
         """Returns True if some ancestor (parent, parent of parent, etc.) of of this location has the given sequence
         type, or False otherwise."""
         if not self.parent:
             return False
         return self.parent.has_ancestor_of_type(sequence_type, include_self=True)
 
-    def lift_over_to_first_ancestor_of_type(self, sequence_type: str) -> "Location":
+    def lift_over_to_first_ancestor_of_type(self, sequence_type: Union[str, SequenceType]) -> "Location":
         """Returns a new Location representing the liftover of this Location to its closest ancestor sequence (parent,
         parent of parent, etc.) which has the given sequence type. If the immediate parent has the given type,
         returns this Location. Raises NoSuchAncestorException if no ancestor with the given type exists."""
@@ -313,7 +337,9 @@ class Location(ABC):
         return lifted_to_grandparent.lift_over_to_sequence(sequence)
 
     @abstractmethod
-    def intersection(self, other: "Location", match_strand: bool = True) -> "Location":
+    def intersection(
+        self, other: "Location", match_strand: bool = True, full_span: bool = False, strict_parent_compare: bool = False
+    ) -> "Location":
         """Returns a new Location representing the intersection of this Location with the other Location.
         Returned Location, if nonempty, has the same Strand as this Location. This operation is commutative
         if match_strand is True.
@@ -325,15 +351,26 @@ class Location(ABC):
         match_strand
             If set to True, automatically return EmptyLocation() if other Location has a different Strand than
             this Location
+        full_span
+            If set to True, compare the full span of this Location to the full span of the other Location.
+        strict_parent_compare
+            Raise MismatchedParentException if parents do not match
+
         """
 
     @abstractmethod
     def union(self, other: "Location") -> "Location":
         """Returns a new Location representing the union of this Location with the other Location. This operation
-        is commutative. Raises ValueError if locations cannot be combined."""
+        is commutative. Raises exception if locations cannot be combined."""
 
     @abstractmethod
-    def minus(self, other: "Location", match_strand: bool = True) -> "Location":
+    def union_preserve_overlaps(self, other: "Location") -> "Location":
+        """Returns a new Location representing the union of this Location with the other Location, retaining
+        overlapping blocks where applicable. This operation is commutative. Raises exception if locations cannot
+        be combined."""
+
+    @abstractmethod
+    def minus(self, other: "Location", match_strand: bool = True, strict_parent_compare: bool = False) -> "Location":
         """Returns a new Location representing this Location minus its intersection with the other Location.
         Returned Location has the same Strand as this Location. If there is no intersection, returns this Location.
         This operation is not commutative.
@@ -345,6 +382,9 @@ class Location(ABC):
         match_strand
             If set to True, automatically return this Location if other Location has a different Strand than this
             Location
+        strict_parent_compare
+            Raise MismatchedParentException if parents do not match
+
         """
 
     @abstractmethod
@@ -372,11 +412,37 @@ class Location(ABC):
             Non-negative integer: amount to extend downstream relative to Strand
         """
 
-    def contains(self, other: "Location", match_strand: bool = False):
-        """Returns True iff this location contains the other"""
-        if not self.has_overlap(other, match_strand):
+    def contains(
+        self,
+        other: "Location",
+        match_strand: bool = False,
+        full_span: bool = False,
+        strict_parent_compare: bool = False,
+    ) -> bool:
+        """Returns True iff this location contains the other. If ``full_span`` is ``True``, the full span of
+        both locations are compared.
+
+        Parameters
+        ----------
+        other
+            Other location
+        match_strand
+            If set to True, automatically return EmptyLocation() if other Location has a different Strand than
+            this Location
+        full_span
+            If set to True, compare the full span of this Location to the full span of the other Location.
+        strict_parent_compare
+            Raise MismatchedParentException if parents do not match
+
+        """
+        if strict_parent_compare:
+            ObjectValidation.require_parents_equal_except_location(self.parent, other.parent)
+        if not self.has_overlap(other, match_strand, full_span):
             return False
         other_to_compare = other.reset_parent(None)
-        return len(self.reset_parent(None).intersection(other_to_compare, match_strand=match_strand)) == len(
-            other_to_compare
-        )
+        if full_span is False:
+            return len(
+                self.reset_parent(None).intersection(other_to_compare, match_strand=match_strand, full_span=full_span)
+            ) == len(other_to_compare)
+        else:
+            return self.reset_parent(None)._full_span_interval.contains(other_to_compare._full_span_interval)
